import numpy as np

import tensorflow as tf
from tensorflow.keras.losses import KLD, MSE


def grad_MainModel(model, input_, target, target_p, total_loss, LVG_target = None, aeloss_fun = None, clust_weight = 1.):
    """Function to do a backprop update to the main CarDEC model for a minibatch.
    
    
    Arguments:
    ------------------------------------------------------------------
    - model: `tensorflow.keras.Model`, The main CarDEC model.
    - input_: `list`, A list containing the input HVG and (optionally) LVG expression tensors of the minibatch for the CarDEC model.
    - target: `tf.Tensor`, Tensor containing the reconstruction target of the minibatch for the HVGs.
    - target_p: `tf.Tensor`, Tensor containing cluster membership probability targets for the minibatch.
    - total_loss: `function`, Function to compute the loss for the main CarDEC model for a minibatch.
    - LVG_target: `tf.Tensor` (Optional), Tensor containing the reconstruction target of the minibatch for the LVGs.
    - aeloss_fun: `function`, Function to compute reconstruction loss.
    - clust_weight: `float`, A float between 0 and 2 balancing clustering and reconstruction losses.
    
    Returns:
    ------------------------------------------------------------------
    - loss_value: `tf.Tensor`: The loss computed for the minibatch.
    - gradients: `a list of Tensors`: Gradients to update the model weights.
    """
    
    with tf.GradientTape() as tape:
        denoised_output, cluster_output = model(*input_)
        loss_value, aeloss = total_loss(target, denoised_output, target_p, cluster_output, 
                                LVG_target, aeloss_fun, clust_weight)
        
    return loss_value, tape.gradient(loss_value, model.trainable_variables)


def grad_reconstruction(model, input_, target, loss):
    """Function to compute gradient update for pretrained autoencoder only.
    
    
    Arguments:
    ------------------------------------------------------------------
    - model: `tensorflow.keras.Model`, The main CarDEC model.
    - input_: `list`, A list containing the input HVG expression tensor of the minibatch for the CarDEC model.
    - target: `tf.Tensor`, Tensor containing the reconstruction target of the minibatch for the HVGs.
    - loss: `function`, Function to compute reconstruction loss.
    
    Returns:
    ------------------------------------------------------------------
    - loss_value: `tf.Tensor`: The loss computed for the minibatch.
    - gradients: `a list of Tensors`: Gradients to update the model weights.
    """
    
    if type(input_) != tuple:
        input_ = (input_, )
        
    with tf.GradientTape() as tape:
        output = model(*input_)
        loss_value = loss(target, output)
        
    return loss_value, tape.gradient(loss_value, model.trainable_variables)


def total_loss(target, denoised_output, p, cluster_output_q, LVG_target = None, aeloss_fun = None, clust_weight = 1.):
    """Function to compute the loss for the main CarDEC model for a minibatch.
    
    
    Arguments:
    ------------------------------------------------------------------
    - target: `tf.Tensor`, Tensor containing the reconstruction target of the minibatch for the HVGs.
    - denoised_output: `dict`, Dictionary containing the output tensors from the CarDEC main model's forward pass.
    - p: `tf.Tensor`, Tensor of shape (n_obs, n_cluster) containing cluster membership probability targets for the minibatch.
    - cluster_output_q: `tf.Tensor`, Tensor of shape (n_obs, n_cluster) containing predicted cluster membership probabilities
    for each cell.
    - LVG_target: `tf.Tensor` (Optional), Tensor containing the reconstruction target of the minibatch for the LVGs.
    - aeloss_fun: `function`, Function to compute reconstruction loss.
    - clust_weight: `float`, A float between 0 and 2 balancing clustering and reconstruction losses.
    
    Returns:
    ------------------------------------------------------------------
    - net_loss: `tf.Tensor`, The loss computed for the minibatch.
    - aeloss: `tf.Tensor`, The reconstruction loss computed for the minibatch.
    """

    if aeloss_fun is not None:
        
        aeloss_HVG = aeloss_fun(target, denoised_output['HVG_denoised'])
        if LVG_target is not None:
            aeloss_LVG = aeloss_fun(LVG_target, denoised_output['LVG_denoised'])
            aeloss = 0.5*(aeloss_LVG + aeloss_HVG)
        else:
            aeloss = 1. * aeloss_HVG
    else:
        aeloss = 0.
    
    net_loss = clust_weight * tf.reduce_mean(KLD(p, cluster_output_q)) + (2. - clust_weight) * aeloss
    
    return net_loss, aeloss


def MSEloss(netinput, netoutput):
    """Function to compute the MSEloss for the reconstruction loss of a minibatch.
    
    
    Arguments:
    ------------------------------------------------------------------
    - netinput: `tf.Tensor`, Tensor containing the network reconstruction target of the minibatch for the cells.
    - netoutput: `tf.Tensor`, Tensor containing the reconstructed target of the minibatch for the cells.
    
    Returns:
    ------------------------------------------------------------------
    - mse_loss: `tf.Tensor`, The loss computed for the minibatch, averaged over genes and cells.
    """
    
    return tf.math.reduce_mean(MSE(netinput, netoutput))


def NBloss(count, output, eps = 1e-10, mean = True):
    """Function to compute the negative binomial reconstruction loss of a minibatch.
    
    
    Arguments:
    ------------------------------------------------------------------
    - count: `tf.Tensor`, Tensor containing the network reconstruction target of the minibatch for the cells (the original 
    counts).
    - output: `tf.Tensor`, Tensor containing the reconstructed target of the minibatch for the cells.
    - eps: `float`, A small number introduced for computational stability
    - mean: `bool`, If True, average negative binomial loss over genes and cells
    
    Returns:
    ------------------------------------------------------------------
    - nbloss: `tf.Tensor`, The loss computed for the minibatch. If mean was True, it has shape (n_obs, n_var). Otherwise, it has shape (1,).
    """
    
    count = tf.cast(count, tf.float32)
    mu = tf.cast(output[0], tf.float32)

    theta = tf.minimum(output[1], 1e6)

    t1 = tf.math.lgamma(theta + eps) + tf.math.lgamma(count + 1.0) - tf.math.lgamma(count + theta + eps)
    t2 = (theta + count) * tf.math.log(1.0 + (mu/(theta+eps))) + (count * (tf.math.log(theta + eps) - tf.math.log(mu + eps)))

    final = _nan2inf(t1 + t2)
    
    if mean:
        final = tf.reduce_sum(final)/final.shape[0]/final.shape[1]

    return final


def ZINBloss(count, output, eps = 1e-10):
    """Function to compute the negative binomial reconstruction loss of a minibatch.
    
    
    Arguments:
    ------------------------------------------------------------------
    - count: `tf.Tensor`, Tensor containing the network reconstruction target of the minibatch for the cells (the original counts).
    - output: `tf.Tensor`, Tensor containing the reconstructed target of the minibatch for the cells.
    - eps: `float`, A small number introduced for computational stability
    
    Returns:
    ------------------------------------------------------------------
    - zinbloss: `tf.Tensor`, The loss computed for the minibatch. Has shape (1,).
    """
    
    mu = output[0]
    theta = output[1]
    pi = output[2]
    
    NB = NBloss(count, output, eps = eps, mean = False) - tf.math.log(1.0 - pi + eps)
    
    count = tf.cast(count, tf.float32)
    mu = tf.cast(mu, tf.float32)
    
    theta = tf.math.minimum(theta, 1e6)
    
    zero_nb = tf.math.pow(theta/(theta + mu + eps), theta)
    zero_case = -tf.math.log(pi + ((1.0- pi) * zero_nb) + eps)
    final = tf.where(tf.less(count, 1e-8), zero_case, NB)
    
    final = tf.reduce_sum(final)/final.shape[0]/final.shape[1]
            
    return final


def _nan2inf(x):
    """Function to replace nan entries in a Tensor with infinities.
    
    
    Arguments:
    ------------------------------------------------------------------
    - x: `tf.Tensor`, Tensor of arbitrary shape.
    
    Returns:
    ------------------------------------------------------------------
    - x': `tf.Tensor`, Tensor x with nan entries replaced by infinity.
    """
    
    return tf.where(tf.math.is_nan(x), tf.zeros_like(x) + np.inf, x)




from .CarDEC_SAE import SAE
from .CarDEC_utils import build_dir, find_resolution
from .CarDEC_layers import ClusteringLayer
from .CarDEC_optimization import grad_MainModel as grad, total_loss, MSEloss

import tensorflow as tf
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Dense, concatenate
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.backend import set_floatx

from sklearn.cluster import KMeans

import scanpy as sc
from anndata import AnnData
import pandas as pd

import random
import numpy as np
from math import ceil

import os
from copy import deepcopy
from time import time

set_floatx('float32')

class CarDEC_Model(Model):
    def __init__(self, adata, dims, LVG_dims = None, tol = 0.005, n_clusters = None, random_seed = 201809, 
                 louvain_seed = 0, n_neighbors = 15, pretrain_epochs = 300, batch_size = 64, decay_factor = 1/3, 
                 patience_LR = 3, patience_ES = 9, act = 'relu', actincenter = "tanh", ae_lr = 1e-04, clust_weight = 1., 
                 load_encoder_weights = True, set_centroids = True, weights_dir = "CarDEC Weights"):
        super(CarDEC_Model, self).__init__()
        """ This class creates the TensorFlow CarDEC model architecture.


        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, the annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes.
        - dims: `list`, the number of output features for each layer of the HVG encoder. The length of the list determines the number of layers.
        - LVG_dims: `list`, the number of output features for each layer of the LVG encoder. The length of the list determines the number of layers.
        - tol: `float`, stop criterion, clustering procedure will be stopped when the difference ratio between the current iteration and last iteration larger than tol.
        - n_clusters: `int`, The number of clusters into which cells will be grouped.
        - random_seed: `int`, The seed used for random weight intialization.
        - louvain_seed: `int`, The seed used for louvain clustering intialization.
        - n_neighbors: `int`, The number of neighbors used for building the graph needed for louvain clustering.
        - pretrain_epochs: `int`, The maximum number of epochs for pretraining the HVG autoencoder. In practice, early stopping criteria should stop training much earlier.
        - batch_size: `int`, The batch size used for pretraining the HVG autoencoder.
        - decay_factor: `float`, The multiplicative factor by which to decay the learning rate when validation loss is not decreasing.
        - patience_LR: `int`, the number of epochs which the validation loss is allowed to increase before learning rate is decayed when pretraining the autoencoder.
        - patience_ES: `int`, the number of epochs which the validation loss is allowed to increase before training is halted when pretraining the autoencoder.
        - act: `str`, The activation function used for the intermediate layers of CarDEC, other than the bottleneck layer.
        - actincenter: `str`, The activation function used for the bottleneck layer of CarDEC.
        - ae_lr: `float`, The learning rate for pretraining the HVG autoencoder.
        - clust_weight: `float`, a number between 0 and 2 qhich balances the clustering and reconstruction losses.
        - load_encoder_weights: `bool`, If True, the API will try to load the weights for the HVG encoder from the weight directory.
        - set_centroids: `bool`, If True, intialize the centroids by running Louvain's algorithm.
        - weights_dir: `str`, the path in which to save the weights of the CarDEC model.
        ------------------------------------------------------------------
        """
        
        assert clust_weight <= 2. and clust_weight>=0.
        
        tf.keras.backend.clear_session()
                    
        self.dims = dims
        self.LVG_dims = LVG_dims
        self.tol = tol
        self.input_dim = dims[0]  # for clustering layer 
        self.n_stacks = len(self.dims) - 1
        self.n_neighbors = n_neighbors
        self.batch_size = batch_size
        self.random_seed = random_seed
        self.activation = act
        self.actincenter = actincenter
        self.load_encoder_weights = load_encoder_weights
        self.clust_weight = clust_weight
        self.weights_dir = weights_dir
        self.preclust_embedding = None
        
        # set random seed
        random.seed(random_seed)
        np.random.seed(random_seed)
        tf.random.set_seed(random_seed)
        self.splitseed = round(abs(10000*np.random.randn()))
        
        # build the autoencoder
        self.sae = SAE(dims = self.dims, act = self.activation, actincenter = self.actincenter, 
                       random_seed = random_seed, splitseed = self.splitseed, init="glorot_uniform", optimizer = Adam(), 
                       weights_dir = weights_dir)
        
        build_dir(self.weights_dir)
        
        decoder_seed = round(100 * abs(np.random.normal()))
        if load_encoder_weights:
            if os.path.isfile("./" + self.weights_dir + "/pretrained_autoencoder_weights.index"):
                print("Pretrain weight index file detected, loading weights.")
                self.sae.load_autoencoder()
                print("Pretrained high variance autoencoder weights initialized.")
            else:
                print("Pretrain weight index file not detected, pretraining autoencoder weights.\n")
                self.sae.train(adata, lr = ae_lr, num_epochs = pretrain_epochs, 
                               batch_size = batch_size, decay_factor = decay_factor, 
                               patience_LR = patience_LR, patience_ES = patience_ES)
                self.sae.load_autoencoder()
        else:
            print("Pre-training high variance autoencoder.\n")
            self.sae.train(adata, lr = ae_lr, num_epochs = pretrain_epochs, 
                           batch_size = batch_size, decay_factor = decay_factor, 
                           patience_LR = patience_LR, patience_ES = patience_ES)
            self.sae.load_autoencoder()
        
        features = self.sae.embed(adata)
        self.preclust_emb = deepcopy(features)
        self.preclust_denoised = self.sae.denoise(adata, batch_size)
                
        if not set_centroids:
            self.init_centroid = np.zeros((n_clusters, self.dims[-1]), dtype = 'float32')
            self.n_clusters = n_clusters
            self.init_pred = np.zeros((adata.shape[0], dims[-1]))
            
        elif louvain_seed is None:
            print("\nInitializing cluster centroids using K-Means")

            kmeans = KMeans(n_clusters=n_clusters, n_init = 20)
            Y_pred_init = kmeans.fit_predict(features)

            self.init_pred = deepcopy(Y_pred_init)
            self.n_clusters = n_clusters
            self.init_centroid = kmeans.cluster_centers_
            
        else:
            print("\nInitializing cluster centroids using the louvain method.")
            
            n_cells = features.shape[0]
            
            if n_cells > 10**5:
                subset = np.random.choice(range(n_cells), 10**5, replace = False)
                adata0 = AnnData(features[subset])
            else: 
                adata0 = AnnData(features)

            sc.pp.neighbors(adata0, n_neighbors = self.n_neighbors, use_rep="X")
            self.resolution = find_resolution(adata0, n_clusters, louvain_seed)
            adata0 = sc.tl.louvain(adata0, resolution = self.resolution, random_state = louvain_seed, copy = True)

            Y_pred_init = adata0.obs['louvain']
            self.init_pred = np.asarray(Y_pred_init, dtype=int)

            features = pd.DataFrame(adata0.X, index = np.arange(0, adata0.shape[0]))
            Group = pd.Series(self.init_pred, index = np.arange(0, adata0.shape[0]), name="Group")
            Mergefeature = pd.concat([features, Group],axis=1)

            self.init_centroid = np.asarray(Mergefeature.groupby("Group").mean())
            self.n_clusters = self.init_centroid.shape[0]

            print("\n " + str(self.n_clusters) + " clusters detected. \n")
        
        self.encoder = self.sae.encoder
        self.decoder = self.sae.decoder
        
        if LVG_dims is not None:
            n_stacks = len(dims) - 1

            LVG_encoder_layers = []

            for i in range(n_stacks-1):
                LVG_encoder_layers.append(Dense(LVG_dims[i + 1], kernel_initializer = 'glorot_uniform', activation = self.activation, name='encoder%d' % i))

            LVG_encoder_layers.append(Dense(LVG_dims[-1], kernel_initializer = 'glorot_uniform', activation = self.actincenter, name='embedding'))
            self.encoderLVG = Sequential(LVG_encoder_layers, name = 'encoderLVG')

        if LVG_dims is not None:
            decoder_layers = []
            for i in range(self.n_stacks - 1, 0, -1):
                decoder_layers.append(Dense(self.LVG_dims[i], kernel_initializer = 'glorot_uniform', 
                                            activation = self.activation, name='decoderLVG%d' % (i-1)))
                
            decoder_layers.append(Dense(self.LVG_dims[0], activation = 'linear', name='outputLVG'))
            self.decoderLVG = Sequential(decoder_layers, name = 'decoderLVG')
        
        self.clustering_layer = ClusteringLayer(centroids = self.init_centroid, name = 'clustering')
        
        del self.sae
        
        self.construct()
        
    def construct(self, summarize = True):
        """ This class method fully initalizes the TensorFlow model.


        Arguments:
        ------------------------------------------------------------------
        - summarize: `bool`, If True, then print a summary of the model architecture.
        """
        
        x = [tf.zeros(shape = (1, self.dims[0]), dtype=float), None]
        if self.LVG_dims is not None:
            x[1] = tf.zeros(shape = (1, self.LVG_dims[0]), dtype=float)
            
        out = self(*x)
        
        if summarize:
            print("\n-----------------------CarDEC Architecture-----------------------\n")
            self.summary()

            print("\n--------------------Encoder Sub-Architecture--------------------\n")
            self.encoder.summary()
            
            print("\n------------------Base Decoder Sub-Architecture------------------\n")
            self.decoder.summary()

            if self.LVG_dims is not None:
                print("\n------------------LVG Encoder Sub-Architecture------------------\n")
                self.encoderLVG.summary()

                print("\n----------------LVG Base Decoder Sub-Architecture----------------\n")
                self.decoderLVG.summary()

    def call(self, hvg, lvg, denoise = True):
        """ This is the forward pass of the model.
        

        ***Inputs***
            - hvg: `tf.Tensor`, an input tensor of shape (n_obs, n_HVG).
            - lvg: `tf.Tensor`, (Optional) an input tensor of shape (n_obs, n_LVG).
            - denoise: `bool`, (Optional) If True, return denoised expression values for each cell.
            
        ***Outputs***
            - denoised_output: `dict`, (Optional) Dictionary containing denoised tensors.
            - cluster_output: `tf.Tensor`, a tensor of cell cluster membership probabilities of shape (n_obs, m).
        """
        
        hvg = self.encoder(hvg)

        cluster_output = self.clustering_layer(hvg)
        
        if not denoise:
            return cluster_output

        HVG_denoised_output = self.decoder(hvg)
        denoised_output = {'HVG_denoised': HVG_denoised_output}

        if self.LVG_dims is not None:
            lvg = self.encoderLVG(lvg)
            z = concatenate([hvg, lvg], axis=1)

            LVG_denoised_output = self.decoderLVG(z)

            denoised_output['LVG_denoised'] = LVG_denoised_output

        return denoised_output, cluster_output

    @staticmethod
    def target_distribution(q):
        """ Updates target distribution cluster assignment probabilities given CarDEC output.
        
        
        Arguments:
        ------------------------------------------------------------------
        - q: `tf.Tensor`, a tensor of shape (b, m) identifying the probability that each of b cells is in each of the m clusters. Obtained as output from CarDEC.
        
        Returns:
        ------------------------------------------------------------------
        - p: `tf.Tensor`, a tensor of shape (b, m) identifying the pseudo-label probability that each of b cells is in each of the m clusters.
        """
        
        weight = q ** 2 / np.sum(q, axis = 0)
        p = weight.T / np.sum(weight, axis = 1)
        return p.T
    
    def make_generators(self, adata, val_split, batch_size, p, splitseed, newseed):
        """ This class method creates training and validation data generators for the current input data and pseudo labels.
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, the annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes.
        - val_split: `float`, The fraction of cells to be reserved for validation during this step.
        - batch_size: `int`, The batch size used for training the full model.
        - p: `tf.Tensor`, a tensor of shape (b, m) identifying the pseudo-label probability that each of b cells is in each of the m clusters.
        - splitseed: `int`, The seed used to split cells between training and validation. Should be consistent between iterations to ensure the same cells are always used for validation.
        - newseed: `int`, The seed that is set after splitting cells between training and validation. Should be different every iteration so that stochastic operations other than splitting cells between training and validation vary between epochs.
        
        Returns:
        ------------------------------------------------------------------
        - train_dataset: `tf.data.Dataset`, Dataset that returns training examples.
        - val_dataset: `tf.data.Dataset`, Dataset that returns validation examples.
        """
        
        n, num_features = adata.X.shape
        
        if self.LVG_dims is None:
            Xobs_target_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"])
            Xobs_norm_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"])
            p_target_ds = tf.data.Dataset.from_tensor_slices(p)
            
            full_train_dataset = tf.data.Dataset.zip((Xobs_norm_ds, Xobs_target_ds, p_target_ds))
            
        else:
            Xobs_target_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
            Xobs_target_LVGds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'LVG'])
                
            Xobs_norm_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
            Xobs_norm_LVGds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'LVG'])
            p_target_ds = tf.data.Dataset.from_tensor_slices(p)
            
            train_dataset = tf.data.Dataset.zip((Xobs_norm_ds, Xobs_norm_LVGds))
            full_train_dataset = tf.data.Dataset.zip((train_dataset, Xobs_target_ds, Xobs_target_LVGds, p_target_ds))

        tf.random.set_seed(splitseed) #Set the seed so we get same validation split always
        full_train_dataset = full_train_dataset.shuffle(n, reshuffle_each_iteration = False)
        tf.random.set_seed(newseed)

        train_dataset = full_train_dataset.skip(round(val_split * n))
        val_dataset = full_train_dataset.take(round(val_split * n)) 

        train_dataset = train_dataset.shuffle(n - round(val_split * n))
        train_dataset = train_dataset.batch(batch_size)

        val_dataset = val_dataset.shuffle(round(val_split * n))
        val_dataset = val_dataset.batch(batch_size)

        return train_dataset, val_dataset
        
    def train_loop(self, train_dataset):
        """ This class method runs the training loop.
        
        
        Arguments:
        ------------------------------------------------------------------
        - train_dataset: `tf.data.Dataset`, Dataset that returns training examples.
        
        Returns:
        ------------------------------------------------------------------
        - epoch_loss_avg: `float`, The mean training loss for the iteration.
        """

        epoch_loss_avg = tf.keras.metrics.Mean()
        
        if self.LVG_dims is not None:
            for inputs, target, LVG_target, batch_p in train_dataset:
                loss_value, grads = grad(self, inputs, target, batch_p, total_loss = total_loss,
                                         LVG_target = LVG_target, aeloss_fun = MSEloss, 
                                         clust_weight = self.clust_weight)
                self.optimizer.apply_gradients(zip(grads, self.trainable_variables))
                epoch_loss_avg(loss_value)
        else:
            for inputs, target, batch_p in train_dataset:
                loss_value, grads = grad(self, [inputs, None], target, batch_p, total_loss = total_loss, 
                                         LVG_target = None, aeloss_fun = MSEloss, 
                                         clust_weight = self.clust_weight)
                self.optimizer.apply_gradients(zip(grads, self.trainable_variables))
                epoch_loss_avg(loss_value)
                
        return epoch_loss_avg.result()
                
    def validation_loop(self, val_dataset):
        """ This class method runs the validation loop.
        
        
        Arguments:
        ------------------------------------------------------------------
        - val_dataset: `tf.data.Dataset`, Dataset that returns validation examples.
        
        Returns:
        ------------------------------------------------------------------
        - epoch_loss_avg: `float`, The mean validation loss for the iteration (reconstruction + clustering loss)
        - epoch_aeloss_avg_val: `float`, The mean validation reconstruction loss for the iteration
        """

        epoch_loss_avg_val = tf.keras.metrics.Mean()
        epoch_aeloss_avg_val = tf.keras.metrics.Mean()
            
        if self.LVG_dims is not None:
            for inputs, target, LVG_target, batch_p in val_dataset:
                denoised_output, cluster_output = self(*inputs)
                loss_value, aeloss = total_loss(target, denoised_output, batch_p, cluster_output, 
                               LVG_target = LVG_target, aeloss_fun = MSEloss, clust_weight = self.clust_weight)
                epoch_loss_avg_val(loss_value)
                epoch_aeloss_avg_val(aeloss)
        else:
            for inputs, target, batch_p in val_dataset:
                denoised_output, cluster_output = self(*[inputs, None])
                loss_value, aeloss = total_loss(target, denoised_output, batch_p, cluster_output, 
                               LVG_target = None, aeloss_fun = MSEloss, clust_weight= self.clust_weight)
                epoch_loss_avg_val(loss_value)
                epoch_aeloss_avg_val(aeloss)
                
        return epoch_loss_avg_val.result(), epoch_aeloss_avg_val.result()
    
    def package_output(self, adata, init_pred, preclust_denoised, preclust_emb):
        """ This class adds some quantities to the adata object.
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, the annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes.
        - init_pred: `np.ndarray`, the array of initial cluster assignments for each cells, of shape (n_obs,).
        - preclust_denoised: `np.ndarray`, This is the array of feature zscores denoised with the pretrained autoencoder of shape (n_obs, n_vars).
        - preclust_emb: `np.ndarray`, This is the latent embedding from the pretrained autoencoder of shape (n_obs, n_embedding).
        """        
        
        adata.obsm['precluster denoised'] = preclust_denoised
        adata.obsm['precluster embedding'] = preclust_emb
        if adata.shape[0] == init_pred.shape[0]:
            adata.obsm['initial assignments'] = init_pred
    
    def embed(self, adata, batch_size):
        """ This class method can be used to compute the low-dimension embedding for HVG features. 
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes.
        - batch_size: `int`, The batch size for filling the array of low dimension embeddings.
        
        Returns:
        ------------------------------------------------------------------
        - embedding: `np.ndarray`, Array of shape (n_obs, p_embedding) containing the HVG embedding for every cell in the dataset.
        """
        
        input_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
        input_ds = input_ds.batch(batch_size)
        
        embedding = np.zeros((adata.shape[0], self.dims[-1]), dtype = 'float32')
        start = 0

        for x in input_ds:
            end = start + x.shape[0]
            embedding[start:end] = self.encoder(x).numpy()
            start = end
            
        return embedding
    
    def embed_LVG(self, adata, batch_size):
        """ This class method can be used to compute the low-dimension embedding for LVG features. 
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes.
        - batch_size: `int`, The batch size for filling the array of low dimension embeddings.
        
        Returns:
        ------------------------------------------------------------------
        - embedding: `np.ndarray`, Array of shape (n_obs, n_embedding) containing the LVG embedding for every cell in the dataset.
        """
        
        input_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'LVG'])
        input_ds = input_ds.batch(batch_size)

        LVG_embedded = np.zeros((adata.shape[0], self.LVG_dims[-1]), dtype = 'float32')
        start = 0

        for x in input_ds:
            end = start + x.shape[0]
            LVG_embedded[start:end] = self.encoderLVG(x).numpy()
            start = end

        return np.concatenate((adata.obsm['embedding'], LVG_embedded), axis = 1)
    
    def make_outputs(self, adata, batch_size, denoise = True):
        """ This class method can be used to pack all relvant outputs into the adata object after training.
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The annotated data matrix of shape (n_obs, n_vars).
        - batch_size: `int`, The batch size for filling the array of low dimension embeddings.
        - denoise: `bool`, Whether to provide denoised expression values for all cells.
        """
        
        if not denoise:
            input_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
            input_ds = input_ds.batch(batch_size)
            adata.obsm["cluster memberships"] = np.zeros((adata.shape[0], self.n_clusters), dtype = 'float32')
            
            start = 0     
            for x in input_ds:
                q_batch = self(x, None, False)
                end = start + q_batch.shape[0]
                adata.obsm["cluster memberships"][start:end] = q_batch.numpy()
            
                start = end
            
            
        elif self.LVG_dims is not None:
            if not ('embedding' in list(adata.obsm) and 'LVG embedding' in list(adata.obsm)):
                        adata.obsm['embedding'] = self.embed(adata, batch_size)
                        adata.obsm['LVG embedding'] = self.embed_LVG(adata, batch_size)
            hvg_ds = tf.data.Dataset.from_tensor_slices(adata.obsm["embedding"])
            lvg_ds = tf.data.Dataset.from_tensor_slices(adata.obsm["LVG embedding"])
            
            input_ds = tf.data.Dataset.zip((hvg_ds, lvg_ds))
            input_ds = input_ds.batch(batch_size)
            
            adata.obsm["cluster memberships"] = np.zeros((adata.shape[0], self.n_clusters), dtype = 'float32')
            adata.layers["denoised"] = np.zeros(adata.shape, dtype = 'float32')
            
            start = 0     
            for input_ in input_ds:
                denoised_batch = {'HVG_denoised': self.decoder(input_[0]), 'LVG_denoised': self.decoderLVG(input_[1])}
                q_batch = self.clustering_layer(input_[0])
                end = start + q_batch.shape[0]
                
                adata.obsm["cluster memberships"][start:end] = q_batch.numpy()
                adata.layers["denoised"][start:end, adata.var['Variance Type'] == 'HVG'] = denoised_batch['HVG_denoised'].numpy()
                adata.layers["denoised"][start:end, adata.var['Variance Type'] == 'LVG'] = denoised_batch['LVG_denoised'].numpy()
            
                start = end
        
        else:
            if not ('embedding' in list(adata.obsm)):
                adata.obsm['embedding'] = self.embed(adata, batch_size)
                    
            input_ds = tf.data.Dataset.from_tensor_slices(adata.obsm["embedding"])
            input_ds = input_ds.batch(batch_size)
            
            adata.obsm["cluster memberships"] = np.zeros((adata.shape[0], self.n_clusters), dtype = 'float32')
            adata.layers["denoised"] = np.zeros(adata.shape, dtype = 'float32')
            
            start = 0
            
            for input_ in input_ds:
                denoised_batch = {'HVG_denoised': self.decoder(input_)}
                q_batch = self.clustering_layer(input_)
                
                end = start + q_batch.shape[0]
                
                adata.obsm["cluster memberships"][start:end] = q_batch.numpy()
                adata.layers["denoised"][start:end] = denoised_batch['HVG_denoised'].numpy()
                
                start = end
                
    def train(self, adata, batch_size = 64, val_split = 0.1, lr = 1e-04, decay_factor = 1/3,
              iteration_patience_LR = 3, iteration_patience_ES = 6, 
              maxiter = 1e3, epochs_fit = 1, optimizer = Adam(), printperiter = None, denoise = True):
        """ This class method can be used to train the main CarDEC model
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The annotated data matrix of shape (n_obs, n_vars).
        - batch_size: `int`, The batch size used for training the full model.
        - val_split: `float`, The fraction of cells to be reserved for validation during this step.
        - lr: `float`, The learning rate for training the full model.
        - decay_factor: `float`, The multiplicative factor by which to decay the learning rate when validation loss is not decreasing.
        - iteration_patience_LR: `int`, The number of iterations tolerated before decaying the learning rate during which the number of cells that change assignment is less than tol.
        - iteration_patience_ES: `int`, The number of iterations tolerated before stopping training during which the number of cells that change assignment is less than tol.
        - maxiter: `int`, The maximum number of iterations allowed to train the full model. In practice, the model will halt training long before hitting this limit.
        - epochs_fit: `int`, The number of epochs during which to fine-tune weights, before updating the target distribution.
        - optimizer: `tensorflow.python.keras.optimizer_v2`, An instance of a TensorFlow optimizer.
        - printperiter: `int`, Optional integer argument. If specified, denoised values will be returned every printperiter epochs, so that the user can evaluate the progress of denoising as training continues.
        - denoise: `bool`, If True, then denoised expression values are provided for all cells.
        
        Returns:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The updated annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes. Depending on the arguments of the train call, some outputs will be added to adata.
        """
        
        total_start = time()
        seedlist = list(1000*np.random.randn(int(maxiter)))
        seedlist = [abs(int(x)) for x in seedlist]
        
        self.optimizer = optimizer
        self.optimizer.lr = lr
        
        # Begin deep clustering
        y_pred_last = np.ones((adata.shape[0],), dtype = int) * -1.

        min_delta = np.inf
        current_aeloss_val = np.inf
        delta_patience_ES = 0
        delta_patience_LR = 0
        delta_stop = False
        
        self.make_outputs(adata, batch_size, denoise = printperiter is not None)
        
        for ite in range(int(maxiter)):
            
            p = self.target_distribution(adata.obsm['cluster memberships'])
            
            train_dataset, val_dataset = self.make_generators(adata, val_split = 0.1, batch_size = batch_size, 
                                                         p = p, splitseed = self.splitseed, newseed = seedlist[ite])

            best_loss = np.inf
            iter_start = time()
                        
            for epoch in range(epochs_fit):
                current_loss_train = self.train_loop(train_dataset)
                current_loss_val, current_aeloss_val = self.validation_loop(val_dataset)
            
            self.make_outputs(adata, batch_size, denoise = printperiter is not None)
            
            y_pred = np.argmax(adata.obsm['cluster memberships'], axis = 1)
                        
            if printperiter is not None:
                if ite % printperiter == 0 and ite > 0:
                    denoising_filename = os.path.join(CarDEC.weights_dir, '/intermediate_denoising/denoised' + ite)
                    outfile = open(denoising_filename,'wb')
                    pickle.dump(adata.layers["denoised"][:, adata.var['Variance Type'] == 'HVG'], outfile)
                    outfile.close()
                    
                    if self.LVG_dims is not None:
                        denoising_filename = os.path.join(CarDEC.weights_dir, '/intermediate_denoising/denoisedLVG' + ite)
                        outfile = open(denoising_filename,'wb')
                        pickle.dump(adata.layers["denoised"][:, adata.var['Variance Type'] == 'LVG'], outfile)
                        outfile.close()
            
            # check stop criterion
            delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / y_pred.shape[0]
            y_pred_last = deepcopy(y_pred)
            
            current_aeloss_val = current_aeloss_val.numpy()
            current_clustloss_val = (current_loss_val.numpy() - (1 - self.clust_weight) * current_aeloss_val)/self.clust_weight
            print("Iter {:03d} Loss: [Training: {:.3f}, Validation Cluster: {:.3f}, Validation AE: {:.3f}], Label Change: {:.3f}, Time: {:.1f} s".format(ite, current_loss_train.numpy(), current_clustloss_val, current_aeloss_val, delta_label, time() - iter_start))
            
            if current_aeloss_val + 10**(-3) < min_delta:
                min_delta = current_aeloss_val
                delta_patience_ES = 0
                delta_patience_LR = 0
                
            if delta_patience_ES >= iteration_patience_ES:
                delta_stop = True
                
            if delta_patience_LR >= iteration_patience_LR:
                self.optimizer.lr = self.optimizer.lr * decay_factor
                delta_patience_LR = 0
                print("\nDecaying Learning Rate to: " + str(self.optimizer.lr.numpy()))

            delta_patience_ES = delta_patience_ES + 1
            delta_patience_LR = delta_patience_LR + 1
            
            if delta_stop and delta_label < self.tol:
                print('\nAutoencoder_loss ', current_aeloss_val, 'not improving.')
                print('Proportion of Labels Changed: ', delta_label, ' is less than tolerance of ', self.tol)
                print('\nReached tolerance threshold. Stop training.')
                break
                
                        
        y0 = pd.Series(y_pred, dtype='category')
        y0.cat.categories = range(0, len(y0.cat.categories))
        print("\nThe final cluster assignments are:")
        x = y0.value_counts()
        print(x.sort_index(ascending=True))
        
        adata.obsm['embedding'] = self.embed(adata, batch_size)
        if self.LVG_dims is not None:
            adata.obsm['LVG embedding'] = self.embed_LVG(adata, batch_size)
            
        del adata.layers['normalized input']
        
        if denoise:
            self.make_outputs(adata, batch_size, denoise = True)
        
        self.save_weights("./" + self.weights_dir + "/tuned_CarDECweights", save_format='tf')
                   
        print("\nTotal Runtime is " + str(time() - total_start))
                
        print("\nThe CarDEC model is now making inference on the data matrix.")
        
        self.package_output(adata, self.init_pred, self.preclust_denoised, self.preclust_emb)
            
        print("Inference completed, results added.")
        
        return adata
    
    def reload_model(self, adata = None, batch_size = 64, denoise = True):
        """ This class method can be used to load the model's saved weights and redo inference.
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, (Optional) The annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes. If left as None, model weights will be reloaded but inference will not be made.
        - batch_size: `int`, The batch size for filling the array of low dimension embeddings.
        - denoise: `bool`, Whether to provide denoised expression values for all cells.
        
        Returns:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, (Optional) The annotated data matrix of shape (n_obs, n_vars). If an adata object was provided as input, the adata object will be returned with inference outputs added.
        """
        
        if os.path.isfile("./" + self.weights_dir + "/tuned_CarDECweights.index"):
            print("Weight index file detected, loading weights.")
            self.load_weights("./" + self.weights_dir + "/tuned_CarDECweights").expect_partial()
            print("CarDEC Model weights loaded successfully.")
        
            if adata is not None:
                print("\nThe CarDEC model is now making inference on the data matrix.")
                
                adata.obsm['embedding'] = self.embed(adata, batch_size)
                if self.LVG_dims is not None:
                    adata.obsm['LVG embedding'] = self.embed_LVG(adata, batch_size)
                    
                del adata.layers['normalized input']
                
                if denoise:
                    self.make_outputs(adata, batch_size, True)
                
                self.package_output(adata, self.init_pred, self.preclust_denoised, self.preclust_emb)
                
                print("Inference completed, results returned.")
                
                return adata

        else:
            print("\nWeight index file not detected, please call CarDEC_Model.train to learn the weights\n")




from .CarDEC_utils import normalize_scanpy
from .CarDEC_MainModel import CarDEC_Model
from .CarDEC_count_decoder import count_model

import tensorflow as tf
from tensorflow.keras.optimizers import Adam
import numpy as np
from pandas import DataFrame

import os

class CarDEC_API:
    def __init__(self, adata, preprocess=True, weights_dir = "CarDEC Weights", batch_key = None, n_high_var = 2000, LVG = True,
                     normalize_samples = True, log_normalize = True, normalize_features = True):
        """ Main CarDEC API the user can use to conduct batch correction and denoising experiments.


        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, the annotated data matrix of shape (n_obs, n_vars). Rows correspond to cells and columns to genes.
        - preprocess: `bool`, If True, then preprocess the data.
        - weights_dir: `str`, the path in which to save the weights of the CarDEC model.
        - batch_key: `str`, string specifying the name of the column in the observation dataframe which identifies the batch of each cell. If this is left as None, then all cells are assumed to be from one batch.
        - n_high_var: `int`, integer specifying the number of genes to be idntified as highly variable. E.g. if n_high_var = 2000, then the 2000 genes with the highest variance are designated as highly variable.
        - LVG: `bool`, If True, also model LVGs. Otherwise, only model HVGs.
        - normalize_samples: `bool`, If True, normalize expression of each gene in each cell by the sum of expression counts in that cell.
        - log_normalize: `bool`, If True, log transform expression. I.e., compute log(expression + 1) for each gene, cell expression count.
        - normalize_features: `bool`, If True, z-score normalize each gene's expression.
        """
    
        if n_high_var is None:
            n_high_var = None
            LVG = False

        self.weights_dir = weights_dir
        self.LVG = LVG

        self.norm_args = (batch_key, n_high_var, LVG, normalize_samples, log_normalize, normalize_features)

        if preprocess:
            self.dataset = normalize_scanpy(adata, *self.norm_args)
        else:
            assert 'Variance Type' in adata.var.keys()
            assert 'normalized input' in adata.layers
            self.dataset = adata

        self.loaded = False
        self.count_loaded = False

    def build_model(self, load_fullmodel = True, dims = [128, 32], LVG_dims = [128, 32], tol = 0.005, n_clusters = None, 
                    random_seed = 201809, louvain_seed = 0, n_neighbors = 15, pretrain_epochs = 2000, batch_size_pretrain = 64,
                    act = 'relu', actincenter = "tanh", ae_lr = 1e-04, ae_decay_factor = 1/3, ae_patience_LR = 3, 
                    ae_patience_ES = 9, clust_weight = 1., load_encoder_weights = True):
        """ Initializes the main CarDEC model.


        Arguments:
        ------------------------------------------------------------------
        - load_fullmodel: `bool`, If True, the API will try to load the weights for the full model from the weight directory.
        - dims: `list`, the number of output features for each layer of the HVG encoder. The length of the list determines the number of layers.
        - LVG_dims: `list`, the number of output features for each layer of the LVG encoder. The length of the list determines the number of layers.
        - tol: `float`, stop criterion, clustering procedure will be stopped when the difference ratio between the current iteration and last iteration larger than tol.
        - n_clusters: `int`, The number of clusters into which cells will be grouped.
        - random_seed: `int`, The seed used for random weight intialization.
        - louvain_seed: `int`, The seed used for louvain clustering intialization.
        - n_neighbors: `int`, The number of neighbors used for building the graph needed for louvain clustering.
        - pretrain_epochs: `int`, The maximum number of epochs for pretraining the HVG autoencoder. In practice, early stopping criteria should stop training much earlier.
        - batch_size_pretrain: `int`, The batch size used for pretraining the HVG autoencoder.
        - act: `str`, The activation function used for the intermediate layers of CarDEC, other than the bottleneck layer.
        - actincenter: `str`, The activation function used for the bottleneck layer of CarDEC.
        - ae_lr: `float`, The learning rate for pretraining the HVG autoencoder.
        - ae_decay_factor: `float`, The multiplicative factor by which to decay the learning rate when validation loss is not decreasing.
        - ae_patience_LR: `int`, the number of epochs which the validation loss is allowed to increase before learning rate is decayed when pretraining the autoencoder.
        - ae_patience_ES: `int`, the number of epochs which the validation loss is allowed to increase before training is halted when pretraining the autoencoder.
        - clust_weight: `float`, a number between 0 and 2 qhich balances the clustering and reconstruction losses.
        - load_encoder_weights: `bool`, If True, the API will try to load the weights for the HVG encoder from the weight directory.
        """
        
        assert n_clusters is not None
        
        if 'normalized input' not in list(self.dataset.layers):
            self.dataset = normalize_scanpy(self.dataset, *self.norm_args)
        
        p = sum(self.dataset.var["Variance Type"] == 'HVG')
        self.dims = [p] + dims
        
        if self.LVG:
            LVG_p = sum(self.dataset.var["Variance Type"] == 'LVG')
            self.LVG_dims = [LVG_p] + LVG_dims
        else:
            self.LVG_dims = None
        
        self.load_fullmodel = load_fullmodel
        self.weights_exist = os.path.isfile("./" + self.weights_dir + "/tuned_CarDECweights.index")
        
        set_centroids = not (self.load_fullmodel and self.weights_exist)
        
        self.model = CarDEC_Model(self.dataset, self.dims, self.LVG_dims, tol, n_clusters, random_seed, louvain_seed, 
                                  n_neighbors, pretrain_epochs, batch_size_pretrain, ae_decay_factor, 
                                  ae_patience_LR, ae_patience_ES, act, actincenter, ae_lr, 
                                  clust_weight, load_encoder_weights, set_centroids, self.weights_dir)
        
    def make_inference(self, batch_size = 64, val_split = 0.1, lr = 1e-04, decay_factor = 1/3,
                       iteration_patience_LR = 3, iteration_patience_ES = 6, maxiter = 1e3, epochs_fit = 1, 
                       optimizer = Adam(), printperiter = None, denoise_all = True, denoise_list = None):
        """ This class method makes inference on the data (batch correction + denoising) with the main CarDEC model


        Arguments:
        ------------------------------------------------------------------
        - batch_size: `int`, The batch size used for training the full model.
        - val_split: `float`, The fraction of cells to be reserved for validation during this step.
        - lr: `float`, The learning rate for training the full model.
        - decay_factor: `float`, The multiplicative factor by which to decay the learning rate when validation loss is not decreasing.
        - iteration_patience_LR: `int`, The number of iterations tolerated before decaying the learning rate during which the number of cells that change assignment is less than tol.
        - iteration_patience_ES: `int`, The number of iterations tolerated before stopping training during which the number of cells that change assignment is less than tol.
        - maxiter: `int`, The maximum number of iterations allowed to train the full model. In practice, the model will halt training long before hitting this limit.
        - epochs_fit: `int`, The number of epochs during which to fine-tune weights, before updating the target distribution.
        - optimizer: `tensorflow.python.keras.optimizer_v2`, An instance of a TensorFlow optimizer.
        - printperiter: `int`, Optional integer argument. If specified, denoised values will be returned every printperiter epochs, so that the user can evaluate the progress of denoising as training continues.
        - denoise_all: `bool`, If True, then denoised expression values are provided for all cells.
        - denoise_list: `list`, An optional list of cell names (as strings). If provided, denoised values will be computed only for cells in this list.

        Returns:
        ------------------------------------------------------------------
        - denoised: `pd.DataFrame`, (Optional) If denoise_list was specified, then this will be an array of denoised expression provided only for listed cells. If denoise_all was instead specified as True, then denoised expression for all cells will be added as a layer to adata.
        """

        if denoise_list is not None:
            denoise_all = False
            
        if not self.loaded:
            if self.load_fullmodel and self.weights_exist:
                self.dataset = self.model.reload_model(self.dataset, batch_size, denoise_all)

            elif not self.weights_exist:
                print("CarDEC Model Weights not detected. Training full model.\n")
                self.dataset = self.model.train(self.dataset, batch_size, val_split, lr, decay_factor,
                               iteration_patience_LR, iteration_patience_ES, maxiter,
                               epochs_fit, optimizer, printperiter, denoise_all)

            else:
                print("Training full model.\n")
                self.dataset = self.model.train(self.dataset, batch_size, val_split, lr, decay_factor, 
                                                iteration_patience_LR, iteration_patience_ES, 
                                                maxiter, epochs_fit, optimizer, printperiter, denoise_all)
            
            
            self.loaded = True
            
        elif denoise_all:
            self.dataset = self.model.make_outputs(self.dataset, batch_size, True)
            
        if denoise_list is not None:
            denoise_list = list(denoise_list)
            indices = [x in denoise_list for x in self.dataset.obs.index]
            denoised = DataFrame(np.zeros((len(denoise_list), self.dataset.shape[1]), dtype = 'float32'))
            denoised.index = self.dataset.obs.index[indices]
            denoised.columns = self.dataset.var.index
            
            
            if self.LVG:
                hvg_ds = tf.data.Dataset.from_tensor_slices(self.dataset.obsm["embedding"][indices])
                lvg_ds = tf.data.Dataset.from_tensor_slices(self.dataset.obsm["LVG embedding"][indices])
            
                input_ds = tf.data.Dataset.zip((hvg_ds, lvg_ds))
                input_ds = input_ds.batch(batch_size)

                start = 0     
                for x in input_ds:
                    denoised_batch = {'HVG_denoised': self.model.decoder(x[0]), 'LVG_denoised': self.model.decoderLVG(x[1])}
                    q_batch = self.model.clustering_layer(x[0])
                    end = start + q_batch.shape[0]

                    denoised.iloc[start:end, np.where(self.dataset.var['Variance Type'] == 'HVG')[0]] = denoised_batch['HVG_denoised'].numpy()
                    denoised.iloc[start:end, np.where(self.dataset.var['Variance Type'] == 'LVG')[0]] = denoised_batch['LVG_denoised'].numpy()

                    start = end

            else:
                input_ds = tf.data.Dataset.from_tensor_slices(self.dataset.obsm["embedding"])

                input_ds = input_ds.batch(batch_size)

                start = 0

                for x in input_ds:
                    denoised_batch = {'HVG_denoised': self.model.decoder(x)}
                    q_batch = self.model.clustering_layer(x)
                    end = start + q_batch.shape[0]

                    denoised.iloc[start:end] = denoised_batch['HVG_denoised'].numpy()

                    start = end
            
            return denoised
            
        print(" ")
            
    def model_counts(self, load_weights = True, act = 'relu', random_seed = 201809,
                     optimizer = Adam(), keep_dispersion = False, num_epochs = 2000, batch_size_count = 64,
                     val_split = 0.1, lr = 1e-03, decay_factor = 1/3, patience_LR = 3, patience_ES = 9, 
                     denoise_all = True, denoise_list = None):
        """ This class method makes inference on the data on the count scale.


        Arguments:
        ------------------------------------------------------------------
        - load_weights: `bool`, If true, the API will attempt to load the weights for the count model.
        - act: `str`, A string specifying the activation function for intermediate layers of the count models.
        - random_seed: `int`, A seed used for weight initialization.
        - optimizer: `tensorflow.python.keras.optimizer_v2`, An instance of a TensorFlow optimizer.
	- keep_dispersion: `bool`, If True, the gene, cell dispersions will be returned as well.
        - num_epochs: `int`, The maximum number of epochs allowed to train each count model. In practice, the model will halt
        training long before hitting this limit.
        - batch_size_count: `int`, The batch size used for training the count models.
        - val_split: `float`, The fraction of cells to be reserved for validation during this step.
        - lr: `float`, The learning rate for training the count models.
        - decay_factor: `float`, The multiplicative factor by which to decay the learning rate when validation loss is not decreasing.
        - patience_LR: `int`, The number of epochs tolerated before decaying the learning rate during which the validation loss does not decrease.
        - patience_ES: `int`, The number of iterations tolerated before stopping training during which the validation loss does not decrease.
        - denoise_all: `bool`, If True, then denoised expression values are provided for all cells.
        - denoise_list: `list`, An optional list of cell names (as strings). If provided, denoised values will be computed only for cells in this list.

        Returns:
        ------------------------------------------------------------------
        - denoised: `pd.DataFrame`, (Optional) If denoise_list was specified, then this will be an array of denoised expression on the count scale provided only for listed cells. If denoise_all was instead specified as True, then denoised expression for all cells will be added as a layer to adata.
        - denoised_dispersion: `pd.DataFrame`, (Optional) If denoise_list was specified and "keep_dispersion" was set to True, then this will be an array of dispersions from the fitted negative binomial model provided only for listed cells. If denoise_all was instead specified as False, but "keep_dispersion" was still True then dispersions for all cells will be added as a layer to adata.
        """
    
        if denoise_list is not None:
            denoise_all = False
        
        if not self.count_loaded:
            weights_dir = os.path.join(self.weights_dir, 'count weights')
            weight_files_exist = os.path.isfile(weights_dir + "/countmodel_weights_HVG Count.index")
            if self.LVG:
                weight_files_exist = weight_files_exist and os.path.isfile(weights_dir + "/countmodel_weights_LVG Count.index")

            init_args = (act, random_seed, self.model.splitseed, optimizer, weights_dir)
            train_args = (num_epochs, batch_size_count, val_split, lr, decay_factor, patience_LR, patience_ES)

            self.nbmodel = count_model(self.dims, *init_args, n_features = self.dims[-1], mode = 'HVG')

            if load_weights and weight_files_exist:
                print("Weight files for count models detected, loading weights.")
                self.nbmodel.load_model()

            elif load_weights:
                print("Weight files for count models not detected. Training HVG count model.\n")
                self.nbmodel.train(self.dataset, *train_args)

            else:
                print("Training HVG count model.\n")
                self.nbmodel.train(self.dataset, *train_args)

            if self.LVG:
                self.nbmodel_lvg = count_model(self.LVG_dims, *init_args, 
                    n_features = self.dims[-1] + self.LVG_dims[-1], mode = 'LVG')

                if load_weights and weight_files_exist:
                    self.nbmodel_lvg.load_model()
                    print("Count model weights loaded successfully.")

                elif load_weights:
                    print("\n \n \n")
                    print("Training LVG count model.\n")
                    self.nbmodel_lvg.train(self.dataset, *train_args)

                else:
                    print("\n \n \n")
                    print("Training LVG count model.\n")
                    self.nbmodel_lvg.train(self.dataset, *train_args)
            
            self.count_loaded = True
            
        if denoise_all:
            self.nbmodel.denoise(self.dataset, keep_dispersion, batch_size_count)
            if self.LVG:
                self.nbmodel_lvg.denoise(self.dataset, keep_dispersion, batch_size_count)
                
        elif denoise_list is not None:
            denoise_list = list(denoise_list)
            indices = [x in denoise_list for x in self.dataset.obs.index]
            denoised = DataFrame(np.zeros((len(denoise_list), self.dataset.shape[1]), dtype = 'float32'))
            denoised.index = self.dataset.obs.index[indices]
            denoised.columns = self.dataset.var.index
            if keep_dispersion:
                denoised_dispersion = DataFrame(np.zeros((len(denoise_list), self.dataset.shape[1]), dtype = 'float32'))
                denoised_dispersion.index = self.dataset.obs.index[indices]
                denoised_dispersion.columns = self.dataset.var.index
            
            input_ds_embed = tf.data.Dataset.from_tensor_slices(self.dataset.obsm['embedding'][indices])
            input_ds_sf = tf.data.Dataset.from_tensor_slices(self.dataset.obs['size factors'][indices])
            input_ds = tf.data.Dataset.zip((input_ds_embed, input_ds_sf))
            input_ds = input_ds.batch(batch_size_count)

            type_indices = np.where(self.dataset.var['Variance Type'] == 'HVG')[0]

            if not keep_dispersion:
                start = 0
                for x in input_ds:
                    end = start + x[0].shape[0]
                    denoised.iloc[start:end, type_indices] = self.nbmodel(*x)[0].numpy()
                    start = end

            else:
                start = 0
                for x in input_ds:
                    end = start + x[0].shape[0]
                    batch_output = self.nbmodel(*x)
                    denoised.iloc[start:end, type_indices] = batch_output[0].numpy()
                    denoised_dispersion.iloc[start:end, type_indices] = batch_output[1].numpy()
                    start = end
            
            if self.LVG:
                input_ds_embed = tf.data.Dataset.from_tensor_slices(self.dataset.obsm['LVG embedding'][indices])
                input_ds_sf = tf.data.Dataset.from_tensor_slices(self.dataset.obs['size factors'][indices])
                input_ds = tf.data.Dataset.zip((input_ds_embed, input_ds_sf))
                input_ds = input_ds.batch(batch_size_count)

                type_indices = np.where(self.dataset.var['Variance Type'] == 'LVG')[0]

                if not keep_dispersion:
                    start = 0
                    for x in input_ds:
                        end = start + x[0].shape[0]
                        denoised.iloc[start:end, type_indices] = self.nbmodel_lvg(*x)[0].numpy()
                        start = end

                else:
                    start = 0
                    for x in input_ds:
                        end = start + x[0].shape[0]
                        batch_output = self.nbmodel_lvg(*x)
                        denoised.iloc[start:end, type_indices] = batch_output[0].numpy()
                        denoised_dispersion.iloc[start:end, type_indices] = batch_output[1].numpy()
                        start = end
                        
            if not keep_dispersion:
                return denoised
            else:
                return denoised, denoised_dispersion




import tensorflow as tf
from tensorflow.keras.layers import Layer

class ClusteringLayer(Layer):
    def __init__(self, centroids = None, n_clusters = None, n_features = None, alpha=1.0, **kwargs):
        """ The clustering layer predicts the a cell's class membership probability for each cell.
        
        
        Arguments:
        ------------------------------------------------------------------
        - centroids: `tf.Tensor`, Initial cluster ceontroids after pretraining the model.
        - n_clusters: `int`, Number of clusters.
        - n_features: `int`, The number of features of the bottleneck embedding space that the centroids live in.
        - alpha: parameter in Student's t-distribution. Default to 1.0.
        """
        
        super(ClusteringLayer, self).__init__(**kwargs)
        self.alpha = alpha
        self.initial_centroids = centroids

        if centroids is not None:
            n_clusters, n_features = centroids.shape

        self.n_features, self.n_clusters = n_features, n_clusters

        assert self.n_clusters is not None
        assert self.n_features is not None

    def build(self, input_shape):
        """ This class method builds the layer fully once it receives an input tensor.
        
        
        Arguments:
        ------------------------------------------------------------------
        - input_shape: `list`, A list specifying the shape of the input tensor.
        """
        
        assert len(input_shape) == 2
        
        self.centroids = self.add_weight(name = 'clusters', shape = (self.n_clusters, self.n_features), initializer = 'glorot_uniform')
        if self.initial_centroids is not None:
            self.set_weights([self.initial_centroids])
            del self.initial_centroids
        
        self.built = True

    def call(self, x, **kwargs):
        """ Forward pass of the clustering layer,
        
        
        ***Inputs***:
            - x: `tf.Tensor`, the embedding tensor of shape = (n_obs, n_var)
        
        ***Returns***:
            - q: `tf.Tensor`, student's t-distribution, or soft labels for each sample of shape = (n_obs, n_clusters)
        """

        q = 1.0 / (1.0 + (tf.reduce_sum(tf.square(tf.expand_dims(x, axis = 1) - self.centroids), axis = 2) / self.alpha))
        q = q**((self.alpha + 1.0) / 2.0)
        q = q / tf.reduce_sum(q, axis = 1, keepdims = True)

        return q

    def compute_output_shape(self, input_shape):
        """ This method infers the output shape from the input shape.
        
        
        Arguments:
        ------------------------------------------------------------------
        - input_shape: `list`, A list specifying the shape of the input tensor.
        
        Returns:
        ------------------------------------------------------------------
        - output_shape: `list`, A tuple specifying the shape of the output for the minibatch (n_obs, n_clusters)
        """
        
        assert input_shape and len(input_shape) == 2
        return input_shape[0], self.n_clusters


from .CarDEC_optimization import grad_reconstruction as grad, MSEloss

import tensorflow as tf
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Dense, concatenate
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.backend import set_floatx
from time import time

import random
import numpy as np
from scipy.stats import zscore
import os


set_floatx('float32')


class SAE(Model):
    def __init__(self, dims, act = 'relu', actincenter = "tanh", 
                 random_seed = 201809, splitseed = 215, init = "glorot_uniform", optimizer = Adam(),
                 weights_dir = 'CarDEC Weights'):
        """ This class method initializes the SAE model.


        Arguments:
        ------------------------------------------------------------------
        - dims: `list`, the number of output features for each layer of the HVG encoder. The length of the list determines the number of layers.
        - act: `str`, The activation function used for the intermediate layers of CarDEC, other than the bottleneck layer.
        - actincenter: `str`, The activation function used for the bottleneck layer of CarDEC.
        - random_seed: `int`, The seed used for random weight intialization.
        - splitseed: `int`, The seed used to split cells between training and validation. Should be consistent between iterations to ensure the same cells are always used for validation.
        - init: `str`, The weight initialization strategy for the autoencoder.
        - optimizer: `tensorflow.python.keras.optimizer_v2`, An instance of a TensorFlow optimizer.
        - weights_dir: `str`, the path in which to save the weights of the CarDEC model.
        """
        
        super(SAE, self).__init__()
        
        tf.keras.backend.clear_session()
        
        self.weights_dir = weights_dir
        
        self.dims = dims
        self.n_stacks = len(dims) - 1
        self.init = init
        self.optimizer = optimizer
        self.random_seed = random_seed
        self.splitseed = splitseed
        
        self.activation = act
        self.actincenter = actincenter #hidden layer activation function
        
        #set random seed
        random.seed(random_seed)
        np.random.seed(random_seed)
        tf.random.set_seed(random_seed)
            
        encoder_layers = []
        for i in range(self.n_stacks-1):
            encoder_layers.append(Dense(self.dims[i + 1], kernel_initializer = self.init, activation = self.activation, name='encoder_%d' % i))
                
        encoder_layers.append(Dense(self.dims[-1], kernel_initializer=self.init, activation=self.actincenter, name='embedding'))
        self.encoder = Sequential(encoder_layers, name = 'encoder')

        decoder_layers = []
        for i in range(self.n_stacks - 1, 0, -1):
            decoder_layers.append(Dense(self.dims[i], kernel_initializer = self.init, activation = self.activation
                                        , name = 'decoder%d' % (i-1)))
            
        decoder_layers.append(Dense(self.dims[0], activation = 'linear', name='output'))
        
        self.decoder = Sequential(decoder_layers, name = 'decoder')
        
        self.construct()

    def call(self, x):
        """ This is the forward pass of the model.
        
        
        ***Inputs***
            - x: `tf.Tensor`, an input tensor of shape (n_obs, p_HVG).
            
        ***Outputs***
            - output: `tf.Tensor`, A (n_obs, p_HVG) tensor of denoised HVG expression.
        """
        
        c = self.encoder(x)

        output = self.decoder(c)
                    
        return output
    
    def load_encoder(self, random_seed = 2312):
        """ This class method can be used to load the encoder weights, while randomly reinitializing the decoder weights.


        Arguments:
        ------------------------------------------------------------------
        - random_seed: `int`, Seed for reinitializing the decoder.
        """
        
        tf.keras.backend.clear_session()
        
        #set random seed
        random.seed(random_seed)
        np.random.seed(random_seed)
        tf.random.set_seed(random_seed)
     
        self.encoder.load_weights("./" + self.weights_dir + "/pretrained_encoder_weights").expect_partial()
        
        decoder_layers = []
        for i in range(self.n_stacks - 1, 0, -1):
            decoder_layers.append(Dense(self.dims[i], kernel_initializer = self.init, activation = self.activation
                                        , name='decoder%d' % (i-1)))
        self.decoder_base = Sequential(decoder_layers, name = 'decoderbase')
        
        self.output_layer = Dense(self.dims[0], activation = 'linear', name='output')
            
        self.construct(summarize = False)
        
    def load_autoencoder(self, ):
        """ This class method can be used to load the full model's weights."""
        
        tf.keras.backend.clear_session()
        
        self.load_weights("./" + self.weights_dir + "/pretrained_autoencoder_weights").expect_partial()
        
    def construct(self, summarize = False):
        """ This class method fully initalizes the TensorFlow model.


        Arguments:
        ------------------------------------------------------------------
        - summarize: `bool`, If True, then print a summary of the model architecture.
        """
        
        x = tf.zeros(shape = (1, self.dims[0]), dtype=float)
        out = self(x)
        
        if summarize:
            print("----------Autoencoder Architecture----------")
            self.summary()

            print("\n----------Encoder Sub-Architecture----------")
            self.encoder.summary()

            print("\n----------Base Decoder Sub-Architecture----------")
            self.decoder.summary()
        
    def denoise(self, adata, batch_size = 64):
        """ This class method can be used to denoise gene expression for each cell.


        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The annotated data matrix of shape (n_obs, n_vars).
        - batch_size: `int`, The batch size used for computing denoised expression.
        
        Returns:
        ------------------------------------------------------------------
        - output: `np.ndarray`, Numpy array of denoised expression of shape (n_obs, n_vars)
        """
        
        input_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
        input_ds = input_ds.batch(batch_size)
        
        output = np.zeros((adata.shape[0], self.dims[0]), dtype = 'float32')
        start = 0
        
        for x in input_ds:
            end = start + x.shape[0]
            output[start:end] = self(x).numpy()
            start = end
        
        return output
        
    def embed(self, adata, batch_size = 64):
        """ This class method can be used to compute the low-dimension embedding for HVG features. 
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The annotated data matrix of shape (n_obs, n_vars).
        - batch_size: `int`, The batch size for filling the array of low dimension embeddings.
        
        Returns:
        ------------------------------------------------------------------
        - embedding: `np.ndarray`, Array of shape (n_obs, n_vars) containing the cell HVG embeddings.
        """
        
        input_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
        input_ds = input_ds.batch(batch_size)
        
        embedding = np.zeros((adata.shape[0], self.dims[-1]), dtype = 'float32')
        start = 0

        for x in input_ds:
            end = start + x.shape[0]
            embedding[start:end] = self.encoder(x).numpy()
            start = end
            
        return embedding
    
    def makegenerators(self, adata, val_split, batch_size, splitseed):
        """ This class method creates training and validation data generators for the current input data.
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, the annotated data matrix of shape (n_obs, n_vars).
        - val_split: `float`, The fraction of cells to be reserved for validation during this step.
        - batch_size: `int`, The batch size used for training the model.
        - splitseed: `int`, The seed used to split cells between training and validation.
        
        Returns:
        ------------------------------------------------------------------
        - train_dataset: `tf.data.Dataset`, Dataset that returns training examples.
        - val_dataset: `tf.data.Dataset`, Dataset that returns validation examples.
        """
        
        n, num_features = adata.X.shape
                
        Xobs_target_ds = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
        train_dataset = tf.data.Dataset.from_tensor_slices(adata.layers["normalized input"][:, adata.var['Variance Type'] == 'HVG'])
        
        full_train_dataset = tf.data.Dataset.zip((train_dataset, Xobs_target_ds))
        
        tf.random.set_seed(splitseed) #Set the seed so we get same validation split always
        full_train_dataset = full_train_dataset.shuffle(n, reshuffle_each_iteration = False)
                
        train_dataset = full_train_dataset.skip(round(val_split * n))
        val_dataset = full_train_dataset.take(round(val_split * n)) 
        
        train_dataset = train_dataset.shuffle(n - round(val_split * n))
        train_dataset = train_dataset.batch(batch_size)
        
        val_dataset = val_dataset.shuffle(round(val_split * n))
        val_dataset = val_dataset.batch(batch_size)
        
        return train_dataset, val_dataset
    
    def train(self, adata, num_epochs = 2000, batch_size = 64, val_split = 0.1, lr = 1e-03, decay_factor = 1/3,
              patience_LR = 3, patience_ES = 9, save_fullmodel = True):
        """ This class method can be used to train the SAE.
        
        
        Arguments:
        ------------------------------------------------------------------
        - adata: `anndata.AnnData`, The annotated data matrix of shape (n_obs, n_vars).
        - num_epochs: `int`, The maximum number of epochs allowed to train the full model. In practice, the model will halt training long before hitting this limit.
        - batch_size: `int`, The batch size used for training the full model.
        - val_split: `float`, The fraction of cells to be reserved for validation during this step.
        - lr: `float`, The learning rate for training the full model.
        - decay_factor: `float`, The multiplicative factor by which to decay the learning rate when validation loss is not decreasing.
        - patience_LR: `int`, The number of epochs tolerated before decaying the learning rate during which the validation loss fails to decrease.
        - patience_ES: `int`, The number of epochs tolerated before stopping training during which the validation loss fails to decrease.
        - save_fullmodel: `bool`, If True, save the full model's weights, not just the encoder.
        """
        
        tf.keras.backend.clear_session()
        
        train_dataset, val_dataset = self.makegenerators(adata, val_split = 0.1, 
                                                         batch_size = batch_size, splitseed = self.splitseed)
        
        counter_LR = 0
        counter_ES = 0
        best_loss = np.inf
        
        self.optimizer.lr = lr
        
        total_start = time()
        for epoch in range(num_epochs):
            epoch_start = time()
            
            epoch_loss_avg = tf.keras.metrics.Mean()
            epoch_loss_avg_val = tf.keras.metrics.Mean()
            
            # Training loop - using batches of batch_size
            for x, target in train_dataset:
                loss_value, grads = grad(self, x, target, MSEloss)
                self.optimizer.apply_gradients(zip(grads, self.trainable_variables))
                epoch_loss_avg(loss_value)  # Add current batch loss
            
            # Validation Loop
            for x, target in val_dataset:
                output = self(x)
                loss_value = MSEloss(target, output)
                epoch_loss_avg_val(loss_value)
            
            current_loss_val = epoch_loss_avg_val.result()

            epoch_time = round(time() - epoch_start, 1)
            
            print("Epoch {:03d}: Training Loss: {:.3f}, Validation Loss: {:.3f}, Time: {:.1f} s".format(epoch, epoch_loss_avg.result().numpy(), epoch_loss_avg_val.result().numpy(), epoch_time))
            
            if(current_loss_val + 10**(-3) < best_loss):
                counter_LR = 0
                counter_ES = 0
                best_loss = current_loss_val
            else:
                counter_LR = counter_LR + 1
                counter_ES = counter_ES + 1

            if patience_ES <= counter_ES:
                break

            if patience_LR <= counter_LR:
                self.optimizer.lr = self.optimizer.lr * decay_factor
                counter_LR = 0
                print("\nDecaying Learning Rate to: " + str(self.optimizer.lr.numpy()))
                
            # End epoch
        
        total_time = round(time() - total_start, 2)
        
        if not os.path.isdir("./" + self.weights_dir):
            os.mkdir("./" + self.weights_dir)
        
        self.save_weights("./" + self.weights_dir + "/pretrained_autoencoder_weights", save_format='tf')
        self.encoder.save_weights("./" + self.weights_dir + "/pretrained_encoder_weights", save_format='tf')
        
        print('\nTraining Completed')
        print("Total training time: " + str(total_time) + " seconds")




#!/usr/bin/env python
# coding: utf-8

# In[ ]:

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cardec",
    version="1.0.2.6",
    author="Justin Lakkis",
    author_email="jlakks@gmail.com",
    description="A deep learning method for joint batch correction, denoting, and clustering of single-cell rna-seq data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jlakkis/CarDEC",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['numpy>=1.18.1', 'pandas>=1.0.1', 'scipy>=1.4.1', 'tensorflow>= 2.0.1, <=2.3.1', 'scikit-learn>=0.22.2.post1', 'scanpy>=1.5.1', 'louvain>=0.6.1'],
    python_requires='>=3.7',
)